% This program calculates the Mean squared filtered residuals of an ARMAX model 
%using the nag routines implemented in  Matlab 5.3
%
%
%the model is 
%       y_{t}=w(B)/d(B) x_(t-b) + theta(B)/phi(B) (1-B)^da n_t
%
%with posibly several inputs x, and where
%
%    w(B)=(w_0-w_1 B^1-...-w_(qx) B^(qx))
%    d(B)=(1-d_1 B^1-...-d_(px) B^(px))
%
%and with
%    n_(t)=c+(1-B)^da (1-B^S)^Da ThetaS(B^S)/PhiS(B^S) theta(B)/phi(B) a_t
%
%where the seasonal part is
%    S:seasonality period (no seasonality is S=0)
%    ThetaS=(1-TS_1 B^S-...-TS_{Q} B^{QxS})
%    PhiS=(1-FS_1 B^S-...-FS_{P} B^{PxS})
%
%and the nonseasonal part is
%    theta(B)=(1-theta_1 B^1-...-theta_q B^q)
%    phi(B)=(1-phi_1 B^1-...-phi_p B^p)
%
%The function is  [Vfil,AIC,BIC,FPE,paramest]=vfilout(Y,X,bx,wx,dx,Qa,Pa,qa,pa,da,Da,S,kfc,estmet,propout)
%
%INPUT:
%            Y: column vector of size Tx1, with T=sample size
%            X: matrix of size TxK, with K the number of exogenous variables
%            bx: vector of size 1xK with the values of b of each regressor x
%            wx=vector of size 1xK with the orders (qx) of w(B) of each regressor (note that order qx means qx+1 parameters, because of w_0)
%            dx=vector of size 1xK with the orders (px) of d(B) of each regressor
%            Qa=scalar with the seasonal order Q of ThetaS(B)
%            Pa=scalar with the seasonal order P of PhiS(B)
%            qa=scalar with the order q of theta(B)
%            pa=scalar with the order pof phi(B)
%            da=order of regular differencing in n_t
%            Da=order of seasonal differencing in n_t
%            S=seasonality period 
%            kfc=1 if c not equal to 0 and will be estimated, and kfc=0 if c restricted to c=0 
%            estmet=estimation methid: 1=least squares  2=exact ML 3=Marginal Likelihood
%            propout=proportion of sample used for estimation
%
%For instance, with two inputs x1 and x2 bx=[bx1 bx2], dx=[dx1 dx2], wx=[wx1 wx2], rx=[rx1 rx2]
%
%
%OUTPUT 
%   Vfil: Mean squared prediction errors of the filtered residuals
%   Vout: Out of sample prediction errors ('recursively icreasing estimation subsample')
%   AIC,BIC,FPE are the usual selection criterion based on in sample residuals
%   paramest=estimated parameters, the order is
%             firstly the ARIMA modelon nt: phi, theta, PhiS,ThetaS
%             then, transfer function w_0,w_1,...,w_(qx),d_1,...,d_(px) for the first input series
%             then, the same for the second input series, and so on
%             and finally the constant c
%
% NOTE: if the selected model has estimation problems, an error message can appear with the legend IFAIL=# .
%       See nag manual for further information (www.nag.com)


function [Vfil,Vout,AIC,BIC,FPE,paramest]=vfilout(Y,X,bx,wx,dx,da,Da,S,Qa,Pa,qa,pa,kfc,estmet,horizon,propout)


kef=estmet;
c=0;
T=length(Y);
npara0=pa+qa+Pa+Qa+sum(dx)+sum(wx)+length(bx)+kfc;
t=T;
y=Y(1:t);
x=X(1:t,:);
rx=2*ones(size(bx));
pe2=zeros(1,horizon);


   %_initial estimation with the whole sample_________________________________
   [parm,res,varefil]=ininagarimax(y,x,pa,da,qa,Pa,Da,Qa,S,bx,dx,wx,rx,kfc,c);
   
   %_final estimation_________________________________________________________
   %kef=1; %1- ls 2- exact ML 3- marginal lik.
   [parm0,res,varin,mr0,mt0]=nagarimax(y,x,pa,da,qa,Pa,Da,Qa,S,bx,dx,wx,rx,parm,kef,kfc); 
   paramest=parm0;
   resin=res(npara0+da:T);
   varml=resin'*resin/T;
   varinc=resin'*resin/(length(resin)-npara0);
   AIC=(T)*log(varml)+2*npara0;
   BIC=(T)*log(varml)+log(T)*npara0; 
   HQ=(T)*log(varml)+3*log(log(T))*npara0;
   FPE=varinc*(1+npara0/(T));

%initial parameters estimates_________________________________________________
esphi0=parm0(1:pa)';
ind0=pa;
estheta0=parm0(ind0+1:ind0+qa)';
ind0=ind0+qa;
esPhiS0=parm0(ind0+1:ind0+Pa)';
ind0=ind0+Pa;
esThetaS0=parm0(ind0+1:ind0+Qa)';
ind0=ind0+Qa;
for in=1:length(bx)
    estw0(in,1:wx(in)+1)=parm0(ind0+1:ind0+wx(in)+1)';
    ind0=ind0+wx(in)+1;
    estd0(in,1:dx(in))=parm0(ind0+1:ind0+dx(in))';
    ind0=ind0+dx(in);
end
estc0=parm0(length(parm0));

ti=npara0+da+Da+S+1;
%parame=zeros(npara0+horizon,T,horizon);
pred2=zeros(T,horizon);
pred1=zeros(T,horizon);
convergen=zeros(T,horizon);
Dt=zeros(T,horizon);
for horiz=1:horizon
 disp('  ')
 disp(['Filtered prediction errors. Horizon=' int2str(horiz)])
 disp('  ')

   for t=ti:T-horiz
      [horiz t];
      disp(['Horizon=' int2str(horiz) '   t=' int2str(t)])
      t;
      Dt=zeros(T,horiz);
      for h=1:horiz;
         Dt(t+h,h)=1;
      end
      %_____ filtering Dt^(T+i)__________________________________________________
      for h=1:horiz
             Z1t(:,h)=filter([1 -estheta0],[1 -esphi0],Dt(:,h));
             Zt(:,h)=filter([1 -esThetaS0],[1 -esPhiS0],Z1t(:,h));
      end
      %estimating the model 
      %y(t)=c+[w1(B)/d1(B)]x1(t-b1)+[w2(B)/d2(B)]x2(t-b2)+...
      %+[wm(B)/dm(B)]xm(t-bm)+W_1 Zt^(T+1)+...+W_horiz Zt^(T+horiz)+[theta(B)/phi(B)]a(t)
      
      bx1=[bx zeros(1,horiz)];
      wx1=[wx zeros(1,horiz)];
      dx1=[dx zeros(1,horiz)];
      rx1=[rx ones(1,horiz)];
      
      y=Y;
      if isempty(bx)==1
         x=[Zt];
      else
         x=[X Zt];
      end
      
      %_initial estimation of the filtered model____
      if t==ti
         [parm,res,varefil]=ininagarimax(y,x,pa,da,qa,Pa,Da,Qa,S,bx1,dx1,wx1,rx1,kfc,estc0);
      end   
      
      %_final estimation_________
      [converg,parm,res,varefil,mr,mt]=filternagarimax(y,x,Dt,horiz,pa,da,qa,Pa,Da,Qa,S,bx1,dx1,wx1,rx1,parm,kef,kfc);
      esphi=parm(1:pa)';
      ind=pa;
      estheta=parm(ind+1:ind+qa)';
      ind=ind+qa;
      esPhiS=parm(ind+1:ind+Pa)';
      ind=ind+Pa;
      esThetaS=parm(ind+1:ind+Qa)';
      ind=ind+Qa;
      for in=1:length(bx)
           estw(in,1:wx1(in)+1)=parm(ind+1:ind+wx1(in)+1)';
           ind=ind+wx1(in)+1;
           estd(in,1:dx1(in))=parm(ind+1:ind+dx1(in))';
           ind=ind+dx1(in);
      end
      W(1:horiz)=parm(ind+1:ind+horiz);
      estc=parm(length(parm));
      parm1=[parm(1:ind0)' estc]';
      
      %_Predictions______________________________________________________________
      oripred=t; %period from where the predictions are made:y_t+h is predicted from t: oripred=t and nfv=h

      %y=Y(1:t+horiz);
      %x=X(1:t+horiz,:);
      y=Y(1:T);
      x=X(1:T,:);



%____arima models for each input________________
%	HERE WE ARE ASSUMING THAT THE INPUT REGRESSORS ARE NONSTOCHASTICS, AND ARE, THEREFORE, KNOWN
% This part should be adapted if we want to use an ARIMA model for the input regressors. See nag instructions

      varinput=zeros(length(bx)+1,1); %it contains the var of residuals of each ARIMA model for the input series
                                      % if the input is non-stochastic it is zero

      sizemodelinput=zeros(7,length(bx)+1); % each column 1,...,m contain the arima orders of each input
	       										 % each column have (p,d,q,P,D,Q,s)' 
                                         % the last column, m+1, is added as workspace
                                        %
      maxsize=max([1 sum(sizemodelinput([1 3 4 6],:))]); %the number of param of the largest input model
      paraminput=zeros(maxsize,length(bx)+1);

      [pred,prederr,stdpred,vare2fil]=prednagarimax(y,x,mr0,mt0,parm1,kfc,horiz,oripred,varinput,sizemodelinput,paraminput);
      pred2(t+horiz,horiz)=prederr(horiz)^2;
      pred1(t+horiz,horiz)=prederr(horiz);
      convergen(t,horiz)=converg;
      pe2(horiz)=pe2(horiz)+prederr(horiz)^2;
  end
  t;
MSPEf(horiz)=(pred2(ti:T-horiz,horiz)'*convergen(ti:T-horiz,horiz))/sum(convergen(ti:T-horiz,horiz));  %only the convegent estimations are used
MSPEfilt(horiz)=pe2(horiz)/(t-ti+1);
end
Vfil=MSPEf;
Vfilt=MSPEfilt;

disp('  ')
disp('Out of sample prediction errors...')
disp('  ')


%and now the classical out-of-sample forecast------------------------------------------------------------------------
%propout=proportion of sample used for estimation

ti=round(T*propout);
pe2=zeros(1,horizon);

for horiz=1:1:horizon
   disp(['Horizon=' int2str(horiz)])
for t=ti:T-horiz
   t;
   y=Y(1:t);
   x=X(1:t,:);
   if t==ti,parm=parm0;end
   [parm,res,vareout,mr,mt]=nagarimax(y,x,pa,da,qa,Pa,Da,Qa,S,bx,dx,wx,rx,parm,kef,kfc); 
  esphi=parm(1:pa)';
  ind=pa;
  estheta=parm(ind+1:ind+qa)';
  ind=ind+qa;
  esPhiS=parm(ind+1:ind+Pa)';
  ind=ind+Pa;
  esThetaS=parm(ind+1:ind+Qa)';
  ind=ind+Qa;
  for in=1:length(bx)
     estw(in,1:wx(in)+1)=parm(ind+1:ind+wx(in)+1)';
     ind=ind+wx(in)+1;
     estd(in,1:dx(in))=parm(ind+1:ind+dx(in))';
     ind=ind+dx(in);
 end
 estc=parm(length(parm));
 
   %__________________________________________________________________________

   %_Predictions______________________________________________________________
   oripred=t; %period from where the predictions are made:y_t+h is predicted from t: oripred=t and nfv=h

   y=Y(1:t+horiz);
   x=X(1:t+horiz,:);


   %____arima models for each input________________
   varinput=zeros(length(bx)+1,1); %it contains the var of residuals of each ARIMA model for the input series
                                   % if the input is non-stochastic it is zero

   sizemodelinput=zeros(7,length(bx)+1); % each column 1,...,m contain the arima orders of each input
	       										 % each column have (p,d,q,P,D,Q,s) 
                                        % the last column, m+1, is added as workspace
   maxsize=max([1 sum(sizemodelinput([1 3 4 6],:))]); %the number pf param of the largest input model
   paraminput=zeros(maxsize,length(bx)+1);

   [pred,prederr,stdpred,vareout2]=prednagarimax(y,x,mr,mt,parm,kfc,horiz,oripred,varinput,sizemodelinput,paraminput);
   
   %pred=predictions of y(oripred+1:oripred+horiz)
   %prederr=y-pred
   %stdpred=estimated std of prediction errors
   %vare2=estimation of var(a_t)
   pe2(horiz)=pe2(horiz)+prederr(horiz)^2;

end
t;
Vout(horiz)=pe2(horiz)/(t-ti+1);
end

%==========================================================================================================

%SUBROUTINES
%==========================================================================================================
%Initial values por arimax using least squares (kef=1)

function [parm,res,vare]=ininagarimax(y,x,pa,da,qa,Pa,Da,Qa,S,bx,dx,wx,rx,kfc,c) 

%bx,px,qx,rx are row vectors with the information of each input series 
%for instance, with two inputs x1 and x2 bx=[bx1 bx2], px=[px1 px2], qx=[qx1 qx2], rx=[rx1 rx2]
%kfc=0; %0= constant held fixed 1= constant is estimated

if isempty(bx)==1 
   ninput=0; 
   xxy=[y]; 
else 
   xxy=[x y];
   nser=size(xxy,2);
   ninput=nser-1; 
end
mr=[pa,da,qa,Pa,Da,Qa,S]; %mr[p,d,q,P,D,Q,s] for the noise component
mt=[bx 0;   %orders of b (lag of input series)
    wx 0;   %qx 
    dx 0;   %px
    rx 0];  
    
%rx=1: single input 2: transfer func. but not pre-observations, they are 0;3: pre-obser. are estimated

para=[];
if pa>0 para=[(0.5/pa)*ones(1,pa)]; end
if qa>0 para=[para (0.1/qa)*ones(1,qa)];  end
if Pa>0 para=[para (0.5/Pa)*ones(1,Pa)];  end
if Qa>0 para=[para (0.1/Qa)*ones(1,Qa)];  end
for in=1:ninput
   if wx(in)>0 para=[para 0.5 (0.5/wx(in))*ones(1,wx(in))]; else para=[para 0.5];  end
   if dx(in)>0 para=[para (0.5/dx(in))*ones(1,dx(in))]; end
end
para=[para c]' ;

kef=1; %1- ls 2- exact ML 3- marginal lik.
nit=500;%number of iterations
kzsp=0; %if =0 use default values of zsp, if neq 0  then read zsp;
zsp=[0.01,1000,2,0.2];
kzef=0; %not to overwrite data
kpriv=0; %not to monitor output
ifail=-1; % To allow for output message of ifail ifail=-1


error=1;
while error>0.001;
[parm,xxy,itc,sd,cm,s,d,ndf,res,sttf,nsttf,zsp,ifailout] = ...
   g13bef(mr,mt,para,xxy,kzsp,zsp,kfc,kpriv,kef,kzef,nit,ifail);
[para parm];
ifailout;
error=max(abs(parm-para));
para=parm;
end

vare=s/(ndf);
return
%================================================================================
%final values por arimax using estmet

function [parm,res,vare,mr,mt]=nagarimax(y,x,pa,da,qa,Pa,Da,Qa,S,bx,dx,wx,rx,param,kef,kfc) 

%bx,px,qx,rx are row vectors with the information of each input series 
%for instance, with two inputs x1 and x2 bx=[bx1 bx2], px=[px1 px2], qx=[qx1 qx2], rx=[rx1 rx2]
%kfc=0= constant held fixed 1= constant is estimated

if isempty(bx)==1 
   ninput=0; 
   xxy=[y]; 
else 
   xxy=[x y];
   nser=size(xxy,2);
   ninput=nser-1; 
end
mr=[pa,da,qa,0,0,0,0]; %mr[p,d,q,P,D,Q,s] for the noise component
mt=[bx 0;  %orders of b (lag of input series)
    wx 0;   %qx 
    dx 0;   %px
    rx 0];  
              % rx=1: single input 2: transfer func. but not pre-observations, they are 0;3: pre-obser. are estimated
para=param; %initial values for parameter estimation
%kef=1; %1- ls 2- exact ML 3- marginal lik.
nit=100;%number of iterations
kzsp=0; %if =0 use default values of zsp, if neq 0  then read zsp;
zsp=[0.01,1000,2,0.2];
kzef=0; %not to overwrite data
kpriv=0; %not to monitor output
ifail=-1; % To allow for output message of ifail ifail=-1


error=1;
totiter=0;
while error>0.001;
[parm,xxy,itc,sd,cm,s,d,ndf,res,sttf,nsttf,zsp,ifailout] = ...
   g13bef(mr,mt,para,xxy,kzsp,zsp,kfc,kpriv,kef,kzef,nit,ifail);
[para parm];
ifailout;
error=max(abs(parm-para));
para=parm;
totiter=totiter+nit;
if totiter>500, error=0, end
end
vare=s/(ndf);
return
%===========================================================================
%filtered estimation

function   [converg,parm,res,vare,mr,mt]=filternagarimax(y,x,Dt,horiz,pa,da,qa,Pa,Da,Qa,S,bx,dx,wx,rx,parm0,kef,kfc);

col=size(x,2);
cold=size(Dt,2); %number of innovative dummies
lx0=col-cold; %number of inputs
bx0=bx(1:lx0);
dx0=dx(1:lx0);
wx0=wx(1:lx0);

%initial parameters_________________________________________________
esphi0=parm0(1:pa)';
ind=pa;
estheta0=parm0(ind+1:ind+qa)';
ind=ind+qa;
esPhiS0=parm0(ind+1:ind+Pa)';
ind=ind+Pa;
esThetaS0=parm0(ind+1:ind+Qa)';
ind=ind+Qa;

for in=1:lx0
    estw0(in,1:wx0(in)+1)=parm0(ind+1:ind+wx0(in)+1)';
    ind=ind+wx0(in)+1;
    estd0(in,1:dx0(in))=parm0(ind+1:ind+dx0(in))';
    ind=ind+dx0(in);
end
estc0=parm0(length(parm0));
for h=1:horiz
      Z1t(:,h)=filter([1 -estheta0],[1 -esphi0],Dt(:,h));
      Zt(:,h)=filter([1 -esThetaS0],[1 -esPhiS0],Z1t(:,h));
end
 
if isempty(bx0)==1
    x=[Zt];
else
    x=[x(:,1:lx0) Zt];
end
 

%___mL_______________________________________________________________

%kef=1; %1- ls 2- exact ML 3- marginal lik.
error=1;
totiter=0;
converg=1;
while error>0.001
    [parm,res,vare,mr,mt]=nagarimax(y,x,pa,da,qa,Pa,Da,Qa,S,bx,dx,wx,rx,parm0,kef,kfc); 
    esphi=parm(1:pa)';
    ind=pa;
    estheta=parm(ind+1:ind+qa)';
    ind=ind+qa;
    esPhiS=parm(ind+1:ind+Pa)';
    ind=ind+Pa;
    esThetaS=parm(ind+1:ind+Qa)';
    ind=ind+Qa;
    for in=1:lx0
        estw(in,1:wx(in)+1)=parm(ind+1:ind+wx(in)+1)';
        ind=ind+wx(in)+1;
        estd(in,1:dx(in))=parm(ind+1:ind+dx(in))';
        ind=ind+dx(in);
    end
    W(1:horiz)=parm(ind+1:ind+horiz);
    estc=parm(length(parm));
    for h=1:horiz
      Z1t(:,h)=filter([1 -estheta],[1 -esphi],Dt(:,h));
      Zt(:,h)=filter([1 -esThetaS],[1 -esPhiS],Z1t(:,h));
    end
    if isempty(bx0)==1
        x=[Zt];
    else
        x=[x(:,1:lx0) Zt];
    end
    error=max(abs(parm-parm0));
    parm0=parm;
    totiter=totiter+1;
    if totiter>100, error=0; converg=0; end  %THIS LINE IS TO AVOID CONVERGENCE PROBLEMS IN ISOLATED POINTS AND THAT CAN HAVE EXESSIVE WEIGHT IN THE MSPE
 end
return
%========================================================================
%prediction with nag

function        [pred,prederr,stdpred,vare2]=prednagarimax(y,x,mr,mt,para,kfc,horiz,oripred,varinput,...
   sizemodelinput,paraminput)

nser=size(sizemodelinput,2);
ninput=nser-1;

if ninput==0
   xxy=[y];
else
   xxy=[x y];
end
nfv=horiz;
nev=oripred;
%predictions are made from t=oripred up to horizon nfv

rmsxy=varinput;
mrx=sizemodelinput;
parx=paraminput;
par=para(1:length(para)-1);
c=para(length(para));

ifail=-1; %-1 is to allow for error messages
kzef=1; %with =0 the y in xxy is replaced with predictions, otherwise xxy is unchanged
if ninput>0
   [xxy,rmsxy,fva,fsd,sttf,nsttf,ifailout] = g13bjf(mr,mt,para,kfc,nev,nfv,xxy,rmsxy,mrx,parx,kzef,ifail);
   vare2=rmsxy(length(rmsxy));
else
   [rms,st,nst,fva,fsd,isf,ifailout] = g13ajf(mr,par,c,y(1:oripred),nfv,kfc,ifail);
   vare2=rms;
end

ifailout;
%fva are the predictions
%fsd are the standard error of predictions

pred=fva;
prederr=y(nev+1:nev+horiz)-pred;
stdpred=fsd;
return
%==================================================================================================
